// ============================================================================
// Rocks'n'Diamonds - McDuffin Strikes Back!
// ----------------------------------------------------------------------------
// (c) 1995-2014 by Artsoft Entertainment
//                  Holger Schemel
//                  info@artsoft.org
//                  https://www.artsoft.org/
// ----------------------------------------------------------------------------
// conf_emp.c
// ============================================================================

// ------- this file was automatically generated -- do not edit by hand -------

#ifndef CONF_EMP_C
#define CONF_EMP_C

// values for graphics configuration (empty elements)

  { "empty_space_1",				"RocksCE.png"		},
  { "empty_space_1.xpos",			"0"			},
  { "empty_space_1.ypos",			"18"			},
  { "empty_space_1.frames",			"1"			},
  { "empty_space_1.EDITOR",			"RocksCE.png"		},
  { "empty_space_1.EDITOR.xpos",		"16"			},
  { "empty_space_1.EDITOR.ypos",		"18"			},

  { "empty_space_2",				"RocksCE.png"		},
  { "empty_space_2.xpos",			"1"			},
  { "empty_space_2.ypos",			"18"			},
  { "empty_space_2.frames",			"1"			},
  { "empty_space_2.EDITOR",			"RocksCE.png"		},
  { "empty_space_2.EDITOR.xpos",		"17"			},
  { "empty_space_2.EDITOR.ypos",		"18"			},

  { "empty_space_3",				"RocksCE.png"		},
  { "empty_space_3.xpos",			"2"			},
  { "empty_space_3.ypos",			"18"			},
  { "empty_space_3.frames",			"1"			},
  { "empty_space_3.EDITOR",			"RocksCE.png"		},
  { "empty_space_3.EDITOR.xpos",		"18"			},
  { "empty_space_3.EDITOR.ypos",		"18"			},

  { "empty_space_4",				"RocksCE.png"		},
  { "empty_space_4.xpos",			"3"			},
  { "empty_space_4.ypos",			"18"			},
  { "empty_space_4.frames",			"1"			},
  { "empty_space_4.EDITOR",			"RocksCE.png"		},
  { "empty_space_4.EDITOR.xpos",		"19"			},
  { "empty_space_4.EDITOR.ypos",		"18"			},

  { "empty_space_5",				"RocksCE.png"		},
  { "empty_space_5.xpos",			"4"			},
  { "empty_space_5.ypos",			"18"			},
  { "empty_space_5.frames",			"1"			},
  { "empty_space_5.EDITOR",			"RocksCE.png"		},
  { "empty_space_5.EDITOR.xpos",		"20"			},
  { "empty_space_5.EDITOR.ypos",		"18"			},

  { "empty_space_6",				"RocksCE.png"		},
  { "empty_space_6.xpos",			"5"			},
  { "empty_space_6.ypos",			"18"			},
  { "empty_space_6.frames",			"1"			},
  { "empty_space_6.EDITOR",			"RocksCE.png"		},
  { "empty_space_6.EDITOR.xpos",		"21"			},
  { "empty_space_6.EDITOR.ypos",		"18"			},

  { "empty_space_7",				"RocksCE.png"		},
  { "empty_space_7.xpos",			"6"			},
  { "empty_space_7.ypos",			"18"			},
  { "empty_space_7.frames",			"1"			},
  { "empty_space_7.EDITOR",			"RocksCE.png"		},
  { "empty_space_7.EDITOR.xpos",		"22"			},
  { "empty_space_7.EDITOR.ypos",		"18"			},

  { "empty_space_8",				"RocksCE.png"		},
  { "empty_space_8.xpos",			"7"			},
  { "empty_space_8.ypos",			"18"			},
  { "empty_space_8.frames",			"1"			},
  { "empty_space_8.EDITOR",			"RocksCE.png"		},
  { "empty_space_8.EDITOR.xpos",		"23"			},
  { "empty_space_8.EDITOR.ypos",		"18"			},

  { "empty_space_9",				"RocksCE.png"		},
  { "empty_space_9.xpos",			"8"			},
  { "empty_space_9.ypos",			"18"			},
  { "empty_space_9.frames",			"1"			},
  { "empty_space_9.EDITOR",			"RocksCE.png"		},
  { "empty_space_9.EDITOR.xpos",		"24"			},
  { "empty_space_9.EDITOR.ypos",		"18"			},

  { "empty_space_10",				"RocksCE.png"		},
  { "empty_space_10.xpos",			"9"			},
  { "empty_space_10.ypos",			"18"			},
  { "empty_space_10.frames",			"1"			},
  { "empty_space_10.EDITOR",			"RocksCE.png"		},
  { "empty_space_10.EDITOR.xpos",		"25"			},
  { "empty_space_10.EDITOR.ypos",		"18"			},

  { "empty_space_11",				"RocksCE.png"		},
  { "empty_space_11.xpos",			"10"			},
  { "empty_space_11.ypos",			"18"			},
  { "empty_space_11.frames",			"1"			},
  { "empty_space_11.EDITOR",			"RocksCE.png"		},
  { "empty_space_11.EDITOR.xpos",		"26"			},
  { "empty_space_11.EDITOR.ypos",		"18"			},

  { "empty_space_12",				"RocksCE.png"		},
  { "empty_space_12.xpos",			"11"			},
  { "empty_space_12.ypos",			"18"			},
  { "empty_space_12.frames",			"1"			},
  { "empty_space_12.EDITOR",			"RocksCE.png"		},
  { "empty_space_12.EDITOR.xpos",		"27"			},
  { "empty_space_12.EDITOR.ypos",		"18"			},

  { "empty_space_13",				"RocksCE.png"		},
  { "empty_space_13.xpos",			"12"			},
  { "empty_space_13.ypos",			"18"			},
  { "empty_space_13.frames",			"1"			},
  { "empty_space_13.EDITOR",			"RocksCE.png"		},
  { "empty_space_13.EDITOR.xpos",		"28"			},
  { "empty_space_13.EDITOR.ypos",		"18"			},

  { "empty_space_14",				"RocksCE.png"		},
  { "empty_space_14.xpos",			"13"			},
  { "empty_space_14.ypos",			"18"			},
  { "empty_space_14.frames",			"1"			},
  { "empty_space_14.EDITOR",			"RocksCE.png"		},
  { "empty_space_14.EDITOR.xpos",		"29"			},
  { "empty_space_14.EDITOR.ypos",		"18"			},

  { "empty_space_15",				"RocksCE.png"		},
  { "empty_space_15.xpos",			"14"			},
  { "empty_space_15.ypos",			"18"			},
  { "empty_space_15.frames",			"1"			},
  { "empty_space_15.EDITOR",			"RocksCE.png"		},
  { "empty_space_15.EDITOR.xpos",		"30"			},
  { "empty_space_15.EDITOR.ypos",		"18"			},

  { "empty_space_16",				"RocksCE.png"		},
  { "empty_space_16.xpos",			"15"			},
  { "empty_space_16.ypos",			"18"			},
  { "empty_space_16.frames",			"1"			},
  { "empty_space_16.EDITOR",			"RocksCE.png"		},
  { "empty_space_16.EDITOR.xpos",		"31"			},
  { "empty_space_16.EDITOR.ypos",		"18"			},


#endif	// CONF_EMP_C
