// ============================================================================
// Rocks'n'Diamonds - McDuffin Strikes Back!
// ----------------------------------------------------------------------------
// (c) 1995-2014 by Artsoft Entertainment
//                  Holger Schemel
//                  info@artsoft.org
//                  https://www.artsoft.org/
// ----------------------------------------------------------------------------
// conf_snd.h
// ============================================================================

// ------- this file was automatically generated -- do not edit by hand -------

#ifndef CONF_SND_H
#define CONF_SND_H

// values for sounds configuration

#define SND_CLASS_DEFAULT_DIGGING			0
#define SND_CLASS_DEFAULT_COLLECTING			1
#define SND_CLASS_DEFAULT_SNAPPING			2
#define SND_CLASS_DEFAULT_PUSHING			3
#define SND_CLASS_DEFAULT_IMPACT			4
#define SND_CLASS_DEFAULT_WALKING			5
#define SND_CLASS_DEFAULT_PASSING			6
#define SND_CLASS_DEFAULT_DYING				7
#define SND_CLASS_DEFAULT_EXPLODING			8
#define SND_CLASS_DEFAULT_HITTING			9
#define SND_CLASS_SP_DEFAULT_EXPLODING			10
#define SND_CLASS_MM_DEFAULT_EXPLODING			11
#define SND_BD_DIAMOND_COLLECTING			12
#define SND_BD_DIAMOND_FALLING				13
#define SND_BD_DIAMOND_IMPACT				14
#define SND_BD_ROCK_PUSHING				15
#define SND_BD_ROCK_FALLING				16
#define SND_BD_ROCK_IMPACT				17
#define SND_BD_MAGIC_WALL_ACTIVATING			18
#define SND_BD_MAGIC_WALL_ACTIVE			19
#define SND_BD_MAGIC_WALL_FILLING			20
#define SND_BD_AMOEBA_WAITING				21
#define SND_BD_AMOEBA_GROWING				22
#define SND_BD_AMOEBA_TURNING_TO_GEM			23
#define SND_BD_AMOEBA_TURNING_TO_ROCK			24
#define SND_BD_BUTTERFLY_MOVING				25
#define SND_BD_BUTTERFLY_WAITING			26
#define SND_BD_FIREFLY_MOVING				27
#define SND_BD_FIREFLY_WAITING				28
#define SND_BDX_GRASS_BALL_FALLING			29
#define SND_BDX_GRASS_BALL_IMPACT			30
#define SND_BDX_GRASS_LOOSE_FALLING			31
#define SND_BDX_GRASS_LOOSE_IMPACT			32
#define SND_BDX_DIAMOND_COLLECTING			33
#define SND_BDX_DIAMOND_FALLING				34
#define SND_BDX_DIAMOND_IMPACT				35
#define SND_BDX_FLYING_DIAMOND_COLLECTING		36
#define SND_BDX_FLYING_DIAMOND_FALLING			37
#define SND_BDX_FLYING_DIAMOND_IMPACT			38
#define SND_BDX_ROCK_PUSHING				39
#define SND_BDX_ROCK_FALLING				40
#define SND_BDX_ROCK_IMPACT				41
#define SND_BDX_FLYING_ROCK_PUSHING			42
#define SND_BDX_FLYING_ROCK_FALLING			43
#define SND_BDX_FLYING_ROCK_IMPACT			44
#define SND_BDX_HEAVY_ROCK_PUSHING			45
#define SND_BDX_HEAVY_ROCK_FALLING			46
#define SND_BDX_HEAVY_ROCK_IMPACT			47
#define SND_BDX_LIGHT_ROCK_PUSHING			48
#define SND_BDX_LIGHT_ROCK_FALLING			49
#define SND_BDX_LIGHT_ROCK_IMPACT			50
#define SND_BDX_WAITING_ROCK_PUSHING			51
#define SND_BDX_CHASING_ROCK_PUSHING			52
#define SND_BDX_NUT_PUSHING				53
#define SND_BDX_NUT_BREAKING				54
#define SND_BDX_NUT_FALLING				55
#define SND_BDX_NUT_IMPACT				56
#define SND_BDX_NITRO_PACK_PUSHING			57
#define SND_BDX_NITRO_PACK_IMPACT			58
#define SND_BDX_BOMB_COLLECTING				59
#define SND_BDX_BOMB_DROPPING				60
#define SND_BDX_MAGIC_WALL_ACTIVE			61
#define SND_BDX_AMOEBA_1_GROWING			62
#define SND_BDX_AMOEBA_1_ACTIVE				63
#define SND_BDX_AMOEBA_1_OTHER				64
#define SND_BDX_AMOEBA_2_GROWING			65
#define SND_BDX_AMOEBA_2_ACTIVE				66
#define SND_BDX_AMOEBA_2_OTHER				67
#define SND_BDX_BUTTERFLY_1_ACTIVE			68
#define SND_BDX_BUTTERFLY_2_ACTIVE			69
#define SND_BDX_FIREFLY_1_ACTIVE			70
#define SND_BDX_FIREFLY_2_ACTIVE			71
#define SND_BDX_DRAGONFLY_ACTIVE			72
#define SND_BDX_STONEFLY_ACTIVE				73
#define SND_BDX_BITER_ACTIVE				74
#define SND_BDX_BITER_EATING				75
#define SND_BDX_COW_ACTIVE				76
#define SND_BDX_PNEUMATIC_HAMMER_ACTIVE			77
#define SND_BDX_COVERED_ACTIVE				78
#define SND_BDX_INBOX_OPENING				79
#define SND_CLASS_BDX_EXIT_OPENING			80
#define SND_BDX_BUBBLE_MOVING				81
#define SND_BDX_BUBBLE_PUSHING				82
#define SND_SP_BASE_DIGGING				83
#define SND_SP_BUGGY_BASE_DIGGING			84
#define SND_SP_BUGGY_BASE_ACTIVE			85
#define SND_SP_INFOTRON_COLLECTING			86
#define SND_SP_INFOTRON_IMPACT				87
#define SND_SP_ZONK_PUSHING				88
#define SND_SP_ZONK_IMPACT				89
#define SND_SP_DISK_RED_COLLECTING			90
#define SND_SP_DISK_ORANGE_PUSHING			91
#define SND_SP_DISK_YELLOW_PUSHING			92
#define SND_CLASS_SP_PORT_PASSING			93
#define SND_CLASS_SP_EXIT_PASSING			94
#define SND_CLASS_SP_EXIT_OPENING			95
#define SND_CLASS_SP_EXIT_CLOSING			96
#define SND_SP_SNIKSNAK_MOVING				97
#define SND_SP_SNIKSNAK_WAITING				98
#define SND_SP_ELECTRON_MOVING				99
#define SND_SP_ELECTRON_WAITING				100
#define SND_SP_TERMINAL_ACTIVATING			101
#define SND_SP_TERMINAL_ACTIVE				102
#define SND_CLASS_SOKOBAN_PUSHING			103
#define SND_CLASS_SOKOBAN_FILLING			104
#define SND_CLASS_SOKOBAN_EMPTYING			105
#define SND_CLASS_PLAYER_MOVING				106
#define SND_SAND_DIGGING				107
#define SND_CLASS_EMERALD_COLLECTING			108
#define SND_CLASS_EMERALD_IMPACT			109
#define SND_DIAMOND_COLLECTING				110
#define SND_DIAMOND_IMPACT				111
#define SND_DIAMOND_BREAKING				112
#define SND_ROCK_PUSHING				113
#define SND_ROCK_IMPACT					114
#define SND_BOMB_PUSHING				115
#define SND_NUT_PUSHING					116
#define SND_NUT_BREAKING				117
#define SND_NUT_IMPACT					118
#define SND_CLASS_DYNAMITE_COLLECTING			119
#define SND_CLASS_DYNAMITE_DROPPING			120
#define SND_CLASS_DYNAMITE_ACTIVE			121
#define SND_CLASS_KEY_COLLECTING			122
#define SND_CLASS_GATE_PASSING				123
#define SND_BUG_MOVING					124
#define SND_BUG_WAITING					125
#define SND_SPACESHIP_MOVING				126
#define SND_SPACESHIP_WAITING				127
#define SND_YAMYAM_MOVING				128
#define SND_YAMYAM_WAITING				129
#define SND_YAMYAM_DIGGING				130
#define SND_ROBOT_MOVING				131
#define SND_ROBOT_WAITING				132
#define SND_ROBOT_WHEEL_ACTIVATING			133
#define SND_ROBOT_WHEEL_ACTIVE				134
#define SND_MAGIC_WALL_ACTIVATING			135
#define SND_MAGIC_WALL_ACTIVE				136
#define SND_MAGIC_WALL_FILLING				137
#define SND_DC_MAGIC_WALL_ACTIVATING			138
#define SND_DC_MAGIC_WALL_ACTIVE			139
#define SND_DC_MAGIC_WALL_FILLING			140
#define SND_CLASS_AMOEBA_WAITING			141
#define SND_CLASS_AMOEBA_GROWING			142
#define SND_CLASS_AMOEBA_DROPPING			143
#define SND_ACID_SPLASHING				144
#define SND_CLASS_QUICKSAND_FILLING			145
#define SND_CLASS_QUICKSAND_EMPTYING			146
#define SND_CLASS_EXIT_OPENING				147
#define SND_CLASS_EXIT_CLOSING				148
#define SND_CLASS_EXIT_PASSING				149
#define SND_CLASS_STEEL_EXIT_OPENING			150
#define SND_CLASS_STEEL_EXIT_CLOSING			151
#define SND_CLASS_STEEL_EXIT_PASSING			152
#define SND_CLASS_EM_EXIT_OPENING			153
#define SND_CLASS_EM_EXIT_CLOSING			154
#define SND_CLASS_EM_EXIT_PASSING			155
#define SND_CLASS_EM_STEEL_EXIT_OPENING			156
#define SND_CLASS_EM_STEEL_EXIT_CLOSING			157
#define SND_CLASS_EM_STEEL_EXIT_PASSING			158
#define SND_PENGUIN_PASSING				159
#define SND_BALLOON_MOVING				160
#define SND_BALLOON_WAITING				161
#define SND_BALLOON_PUSHING				162
#define SND_CLASS_BALLOON_SWITCH_ACTIVATING		163
#define SND_SPRING_MOVING				164
#define SND_SPRING_PUSHING				165
#define SND_SPRING_IMPACT				166
#define SND_CLASS_WALL_GROWING				167
#define SND_EMC_ANDROID_PUSHING				168
#define SND_EMC_ANDROID_MOVING				169
#define SND_EMC_ANDROID_DROPPING			170
#define SND_EMC_MAGIC_BALL_DROPPING			171
#define SND_PEARL_COLLECTING				172
#define SND_PEARL_BREAKING				173
#define SND_PEARL_IMPACT				174
#define SND_CRYSTAL_COLLECTING				175
#define SND_CRYSTAL_IMPACT				176
#define SND_CLASS_ENVELOPE_COLLECTING			177
#define SND_CLASS_ENVELOPE_OPENING			178
#define SND_CLASS_ENVELOPE_CLOSING			179
#define SND_INVISIBLE_SAND_DIGGING			180
#define SND_INVISIBLE_SAND_ACTIVE_DIGGING		181
#define SND_SHIELD_NORMAL_COLLECTING			182
#define SND_SHIELD_NORMAL_ACTIVE			183
#define SND_SHIELD_DEADLY_COLLECTING			184
#define SND_SHIELD_DEADLY_ACTIVE			185
#define SND_EXTRA_TIME_COLLECTING			186
#define SND_MOLE_MOVING					187
#define SND_MOLE_WAITING				188
#define SND_MOLE_DIGGING				189
#define SND_CLASS_SWITCHGATE_SWITCH_ACTIVATING		190
#define SND_CLASS_SWITCHGATE_OPENING			191
#define SND_CLASS_SWITCHGATE_CLOSING			192
#define SND_CLASS_SWITCHGATE_PASSING			193
#define SND_CLASS_TIMEGATE_SWITCH_ACTIVATING		194
#define SND_CLASS_TIMEGATE_SWITCH_ACTIVE		195
#define SND_CLASS_TIMEGATE_SWITCH_DEACTIVATING		196
#define SND_CLASS_TIMEGATE_OPENING			197
#define SND_CLASS_TIMEGATE_CLOSING			198
#define SND_CLASS_TIMEGATE_PASSING			199
#define SND_CLASS_CONVEYOR_BELT_SWITCH_ACTIVATING	200
#define SND_CLASS_CONVEYOR_BELT_ACTIVE			201
#define SND_CLASS_CONVEYOR_BELT_SWITCH_DEACTIVATING	202
#define SND_LIGHT_SWITCH_ACTIVATING			203
#define SND_LIGHT_SWITCH_DEACTIVATING			204
#define SND_DX_SUPABOMB_PUSHING				205
#define SND_TRAP_DIGGING				206
#define SND_TRAP_ACTIVATING				207
#define SND_CLASS_TUBE_WALKING				208
#define SND_AMOEBA_TURNING_TO_GEM			209
#define SND_AMOEBA_TURNING_TO_ROCK			210
#define SND_SPEED_PILL_COLLECTING			211
#define SND_DYNABOMB_INCREASE_NUMBER_COLLECTING		212
#define SND_DYNABOMB_INCREASE_SIZE_COLLECTING		213
#define SND_DYNABOMB_INCREASE_POWER_COLLECTING		214
#define SND_CLASS_DYNABOMB_DROPPING			215
#define SND_CLASS_DYNABOMB_ACTIVE			216
#define SND_SATELLITE_MOVING				217
#define SND_SATELLITE_WAITING				218
#define SND_SATELLITE_PUSHING				219
#define SND_LAMP_ACTIVATING				220
#define SND_LAMP_DEACTIVATING				221
#define SND_TIME_ORB_FULL_COLLECTING			222
#define SND_TIME_ORB_FULL_IMPACT			223
#define SND_TIME_ORB_EMPTY_PUSHING			224
#define SND_TIME_ORB_EMPTY_IMPACT			225
#define SND_GAME_OF_LIFE_WAITING			226
#define SND_GAME_OF_LIFE_GROWING			227
#define SND_BIOMAZE_WAITING				228
#define SND_BIOMAZE_GROWING				229
#define SND_PACMAN_MOVING				230
#define SND_PACMAN_WAITING				231
#define SND_PACMAN_DIGGING				232
#define SND_DARK_YAMYAM_MOVING				233
#define SND_DARK_YAMYAM_WAITING				234
#define SND_DARK_YAMYAM_DIGGING				235
#define SND_PENGUIN_MOVING				236
#define SND_PENGUIN_WAITING				237
#define SND_PIG_MOVING					238
#define SND_PIG_WAITING					239
#define SND_PIG_DIGGING					240
#define SND_DRAGON_MOVING				241
#define SND_DRAGON_WAITING				242
#define SND_DRAGON_ATTACKING				243
#define SND_CLASS_MM_MCDUFFIN_HITTING			244
#define SND_CLASS_MM_MIRROR_HITTING			245
#define SND_CLASS_MM_MIRROR_FIXED_HITTING		246
#define SND_CLASS_MM_PRISM_HITTING			247
#define SND_CLASS_MM_EXIT_HITTING			248
#define SND_CLASS_MM_EXIT_OPENING			249
#define SND_MM_EXIT_OPEN_HITTING			250
#define SND_CLASS_DF_MIRROR_HITTING			251
#define SND_CLASS_DF_MIRROR_ROTATING_HITTING		252
#define SND_CLASS_DF_REFRACTOR_HITTING			253
#define SND_CLASS_DF_RECEIVER_HITTING			254
#define SND_CLASS_DF_RECEIVER_OPENING			255
#define SND_CLASS_MM_WOODEN_WALL_HITTING		256
#define SND_CLASS_MM_WOODEN_BLOCK_HITTING		257
#define SND_CLASS_MM_WOODEN_BLOCK_PUSHING		258
#define SND_CLASS_MM_WOODEN_LOCK_HITTING		259
#define SND_CLASS_MM_WOODEN_GRID_FIXED_HITTING		260
#define SND_CLASS_MM_FUSE_HITTING			261
#define SND_CLASS_MM_ICE_WALL_HITTING			262
#define SND_CLASS_MM_ICE_WALL_SHRINKING			263
#define SND_CLASS_MM_AMOEBA_WALL_HITTING		264
#define SND_CLASS_MM_AMOEBA_WALL_GROWING		265
#define SND_CLASS_DF_WOODEN_WALL_HITTING		266
#define SND_CLASS_DF_WOODEN_GRID_FIXED_HITTING		267
#define SND_CLASS_DF_WOODEN_GRID_ROTATING_HITTING	268
#define SND_CLASS_MM_STEEL_WALL_HITTING			269
#define SND_CLASS_MM_STEEL_GRID_FIXED_HITTING		270
#define SND_CLASS_MM_STEEL_BLOCK_HITTING		271
#define SND_CLASS_MM_STEEL_BLOCK_PUSHING		272
#define SND_CLASS_MM_STEEL_LOCK_HITTING			273
#define SND_CLASS_DF_STEEL_WALL_HITTING			274
#define SND_CLASS_DF_STEEL_GRID_FIXED_HITTING		275
#define SND_CLASS_DF_STEEL_GRID_ROTATING_HITTING	276
#define SND_CLASS_MM_PACMAN_EXPLODING			277
#define SND_CLASS_MM_MCDUFFIN_EXPLODING			278
#define SND_CLASS_MM_BOMB_EXPLODING			279
#define SND_CLASS_MM_KEY_EXPLODING			280
#define SND_CLASS_MM_STEEL_LOCK_EXPLODING		281
#define SND_CLASS_MM_WOODEN_LOCK_EXPLODING		282
#define SND_BDX_EMPTY_SPACE_WALKING_RANDOM_1		283
#define SND_BDX_EMPTY_SPACE_WALKING_RANDOM_2		284
#define SND_BDX_EMPTY_SPACE_WALKING_RANDOM_3		285
#define SND_BDX_EMPTY_SPACE_WALKING_RANDOM_4		286
#define SND_BDX_EMPTY_SPACE_WALKING_RANDOM_5		287
#define SND_BDX_EMPTY_SPACE_WALKING_RANDOM_6		288
#define SND_BDX_EMPTY_SPACE_WALKING_RANDOM_7		289
#define SND_BDX_EMPTY_SPACE_WALKING_RANDOM_8		290
#define SND_BDX_SAND_DIGGING_RANDOM_1			291
#define SND_BDX_SAND_DIGGING_RANDOM_2			292
#define SND_BDX_SAND_DIGGING_RANDOM_3			293
#define SND_BDX_SAND_DIGGING_RANDOM_4			294
#define SND_BDX_SAND_DIGGING_RANDOM_5			295
#define SND_BDX_SAND_DIGGING_RANDOM_6			296
#define SND_BDX_SAND_DIGGING_RANDOM_7			297
#define SND_BDX_SAND_DIGGING_RANDOM_8			298
#define SND_BDX_ROCK_IMPACT_RANDOM_1			299
#define SND_BDX_ROCK_IMPACT_RANDOM_2			300
#define SND_BDX_ROCK_IMPACT_RANDOM_3			301
#define SND_BDX_ROCK_IMPACT_RANDOM_4			302
#define SND_BDX_ROCK_IMPACT_RANDOM_5			303
#define SND_BDX_ROCK_IMPACT_RANDOM_6			304
#define SND_BDX_ROCK_IMPACT_RANDOM_7			305
#define SND_BDX_ROCK_IMPACT_RANDOM_8			306
#define SND_BDX_HEAVY_ROCK_IMPACT_RANDOM_1		307
#define SND_BDX_HEAVY_ROCK_IMPACT_RANDOM_2		308
#define SND_BDX_HEAVY_ROCK_IMPACT_RANDOM_3		309
#define SND_BDX_HEAVY_ROCK_IMPACT_RANDOM_4		310
#define SND_BDX_HEAVY_ROCK_IMPACT_RANDOM_5		311
#define SND_BDX_HEAVY_ROCK_IMPACT_RANDOM_6		312
#define SND_BDX_HEAVY_ROCK_IMPACT_RANDOM_7		313
#define SND_BDX_HEAVY_ROCK_IMPACT_RANDOM_8		314
#define SND_BDX_LIGHT_ROCK_IMPACT_RANDOM_1		315
#define SND_BDX_LIGHT_ROCK_IMPACT_RANDOM_2		316
#define SND_BDX_LIGHT_ROCK_IMPACT_RANDOM_3		317
#define SND_BDX_LIGHT_ROCK_IMPACT_RANDOM_4		318
#define SND_BDX_LIGHT_ROCK_IMPACT_RANDOM_5		319
#define SND_BDX_LIGHT_ROCK_IMPACT_RANDOM_6		320
#define SND_BDX_LIGHT_ROCK_IMPACT_RANDOM_7		321
#define SND_BDX_LIGHT_ROCK_IMPACT_RANDOM_8		322
#define SND_BDX_DIAMOND_FALLING_RANDOM_1		323
#define SND_BDX_DIAMOND_FALLING_RANDOM_2		324
#define SND_BDX_DIAMOND_FALLING_RANDOM_3		325
#define SND_BDX_DIAMOND_FALLING_RANDOM_4		326
#define SND_BDX_DIAMOND_FALLING_RANDOM_5		327
#define SND_BDX_DIAMOND_FALLING_RANDOM_6		328
#define SND_BDX_DIAMOND_FALLING_RANDOM_7		329
#define SND_BDX_DIAMOND_FALLING_RANDOM_8		330
#define SND_BDX_DIAMOND_IMPACT_RANDOM_1			331
#define SND_BDX_DIAMOND_IMPACT_RANDOM_2			332
#define SND_BDX_DIAMOND_IMPACT_RANDOM_3			333
#define SND_BDX_DIAMOND_IMPACT_RANDOM_4			334
#define SND_BDX_DIAMOND_IMPACT_RANDOM_5			335
#define SND_BDX_DIAMOND_IMPACT_RANDOM_6			336
#define SND_BDX_DIAMOND_IMPACT_RANDOM_7			337
#define SND_BDX_DIAMOND_IMPACT_RANDOM_8			338
#define SND_BDX_FLYING_DIAMOND_FALLING_RANDOM_1		339
#define SND_BDX_FLYING_DIAMOND_FALLING_RANDOM_2		340
#define SND_BDX_FLYING_DIAMOND_FALLING_RANDOM_3		341
#define SND_BDX_FLYING_DIAMOND_FALLING_RANDOM_4		342
#define SND_BDX_FLYING_DIAMOND_FALLING_RANDOM_5		343
#define SND_BDX_FLYING_DIAMOND_FALLING_RANDOM_6		344
#define SND_BDX_FLYING_DIAMOND_FALLING_RANDOM_7		345
#define SND_BDX_FLYING_DIAMOND_FALLING_RANDOM_8		346
#define SND_BDX_FLYING_DIAMOND_IMPACT_RANDOM_1		347
#define SND_BDX_FLYING_DIAMOND_IMPACT_RANDOM_2		348
#define SND_BDX_FLYING_DIAMOND_IMPACT_RANDOM_3		349
#define SND_BDX_FLYING_DIAMOND_IMPACT_RANDOM_4		350
#define SND_BDX_FLYING_DIAMOND_IMPACT_RANDOM_5		351
#define SND_BDX_FLYING_DIAMOND_IMPACT_RANDOM_6		352
#define SND_BDX_FLYING_DIAMOND_IMPACT_RANDOM_7		353
#define SND_BDX_FLYING_DIAMOND_IMPACT_RANDOM_8		354
#define SND_GAME_STARTING				355
#define SND_GAME_LEVELTIME_CHARGING			356
#define SND_GAME_HEALTH_CHARGING			357
#define SND_GAME_RUNNING_OUT_OF_TIME			358
#define SND_GAME_RUNNING_OUT_OF_TIME_10			359
#define SND_GAME_RUNNING_OUT_OF_TIME_9			360
#define SND_GAME_RUNNING_OUT_OF_TIME_8			361
#define SND_GAME_RUNNING_OUT_OF_TIME_7			362
#define SND_GAME_RUNNING_OUT_OF_TIME_6			363
#define SND_GAME_RUNNING_OUT_OF_TIME_5			364
#define SND_GAME_RUNNING_OUT_OF_TIME_4			365
#define SND_GAME_RUNNING_OUT_OF_TIME_3			366
#define SND_GAME_RUNNING_OUT_OF_TIME_2			367
#define SND_GAME_RUNNING_OUT_OF_TIME_1			368
#define SND_GAME_RUNNING_OUT_OF_TIME_0			369
#define SND_GAME_LEVELTIME_BONUS			370
#define SND_GAME_HEALTH_BONUS				371
#define SND_GAME_TIMEOUT				372
#define SND_GAME_LOSING					373
#define SND_GAME_WINNING				374
#define SND_GAME_SOKOBAN_SOLVING			375
#define SND_DOOR_OPENING				376
#define SND_DOOR_CLOSING				377
#define SND_DOOR_1_OPENING				378
#define SND_DOOR_1_CLOSING				379
#define SND_DOOR_2_OPENING				380
#define SND_DOOR_2_CLOSING				381
#define SND_REQUEST_OPENING				382
#define SND_REQUEST_CLOSING				383
#define SND_MENU_ITEM_ACTIVATING			384
#define SND_MENU_ITEM_SELECTING				385
#define SND_MENU_BUTTON_PRESSING			386
#define SND_MENU_BUTTON_RELEASING			387
#define SND_BACKGROUND_TITLE_INITIAL			388
#define SND_BACKGROUND_TITLE				389
#define SND_BACKGROUND_MAIN				390
#define SND_BACKGROUND_NAMES				391
#define SND_BACKGROUND_LEVELS				392
#define SND_BACKGROUND_LEVELNR				393
#define SND_BACKGROUND_SCORES				394
#define SND_BACKGROUND_EDITOR				395
#define SND_BACKGROUND_INFO				396
#define SND_BACKGROUND_INFO_ELEMENTS			397
#define SND_BACKGROUND_INFO_CREDITS			398
#define SND_BACKGROUND_INFO_PROGRAM			399
#define SND_BACKGROUND_INFO_VERSION			400
#define SND_BACKGROUND_INFO_LEVELSET			401
#define SND_BACKGROUND_INFO_LEVEL			402
#define SND_BACKGROUND_STORY				403
#define SND_BACKGROUND_SETUP				404
#define SND_BACKGROUND_TITLESCREEN_INITIAL_1		405
#define SND_BACKGROUND_TITLESCREEN_INITIAL_2		406
#define SND_BACKGROUND_TITLESCREEN_INITIAL_3		407
#define SND_BACKGROUND_TITLESCREEN_INITIAL_4		408
#define SND_BACKGROUND_TITLESCREEN_INITIAL_5		409
#define SND_BACKGROUND_TITLESCREEN_1			410
#define SND_BACKGROUND_TITLESCREEN_2			411
#define SND_BACKGROUND_TITLESCREEN_3			412
#define SND_BACKGROUND_TITLESCREEN_4			413
#define SND_BACKGROUND_TITLESCREEN_5			414
#define SND_BACKGROUND_TITLEMESSAGE_INITIAL_1		415
#define SND_BACKGROUND_TITLEMESSAGE_INITIAL_2		416
#define SND_BACKGROUND_TITLEMESSAGE_INITIAL_3		417
#define SND_BACKGROUND_TITLEMESSAGE_INITIAL_4		418
#define SND_BACKGROUND_TITLEMESSAGE_INITIAL_5		419
#define SND_BACKGROUND_TITLEMESSAGE_1			420
#define SND_BACKGROUND_TITLEMESSAGE_2			421
#define SND_BACKGROUND_TITLEMESSAGE_3			422
#define SND_BACKGROUND_TITLEMESSAGE_4			423
#define SND_BACKGROUND_TITLEMESSAGE_5			424

#define NUM_SOUND_FILES					425

#endif	// CONF_SND_H
