// ============================================================================
// Rocks'n'Diamonds - McDuffin Strikes Back!
// ----------------------------------------------------------------------------
// (c) 1995-2014 by Artsoft Entertainment
//                  Holger Schemel
//                  info@artsoft.org
//                  https://www.artsoft.org/
// ----------------------------------------------------------------------------
// conf_emp.h
// ============================================================================

// ------- this file was automatically generated -- do not edit by hand -------

#ifndef CONF_EMP_H
#define CONF_EMP_H

// values for elements configuration (empty elements)

#define EL_EMPTY_SPACE_1		(EL_EMPTY_SPACE_START + 0)
#define EL_EMPTY_SPACE_2		(EL_EMPTY_SPACE_START + 1)
#define EL_EMPTY_SPACE_3		(EL_EMPTY_SPACE_START + 2)
#define EL_EMPTY_SPACE_4		(EL_EMPTY_SPACE_START + 3)
#define EL_EMPTY_SPACE_5		(EL_EMPTY_SPACE_START + 4)
#define EL_EMPTY_SPACE_6		(EL_EMPTY_SPACE_START + 5)
#define EL_EMPTY_SPACE_7		(EL_EMPTY_SPACE_START + 6)
#define EL_EMPTY_SPACE_8		(EL_EMPTY_SPACE_START + 7)
#define EL_EMPTY_SPACE_9		(EL_EMPTY_SPACE_START + 8)
#define EL_EMPTY_SPACE_10		(EL_EMPTY_SPACE_START + 9)
#define EL_EMPTY_SPACE_11		(EL_EMPTY_SPACE_START + 10)
#define EL_EMPTY_SPACE_12		(EL_EMPTY_SPACE_START + 11)
#define EL_EMPTY_SPACE_13		(EL_EMPTY_SPACE_START + 12)
#define EL_EMPTY_SPACE_14		(EL_EMPTY_SPACE_START + 13)
#define EL_EMPTY_SPACE_15		(EL_EMPTY_SPACE_START + 14)
#define EL_EMPTY_SPACE_16		(EL_EMPTY_SPACE_START + 15)

#endif	// CONF_EMP_C
